local ceil = ceil
local pairs = pairs
local select = select
local GetTime = GetTime
local strfind = strfind
local tinsert = tinsert
local tremove = tremove
local tonumber = tonumber
local strsplit = strsplit
local UnitName = UnitName
local UnitRace = UnitRace
local UnitGUID = UnitGUID
local UnitClass = UnitClass
local UnitExists = UnitExists
local GetSpellInfo = GetSpellInfo
local IsInInstance = IsInInstance
local GetNumGroupMembers = GetNumGroupMembers
local EventFrame = CreateFrame("Frame")
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local Font = "Interface\\Addons\\CCIcons\\Prototype.ttf"
local Texture = "Interface\\AddOns\\CCIcons\\Ferous.tga"
local LGI = LibStub:GetLibrary("LibGroupInSpecT-1.1")
local ENABLE_ICONS = true
local ENABLE_COOLDOWNS = true
local CC_ICON_SIZE = 60

-- Spells
local REMORSELESS_WINTER = GetSpellInfo(108200)
local SHOCKWAVE = GetSpellInfo(46968)
local LEG_SWEEP = GetSpellInfo(119381)
local FISTS_OF_FURY = GetSpellInfo(113656)
local CAPACITOR_TOTEM = GetSpellInfo(108269)
local BINDING_SHOT = GetSpellInfo(109248)
local BLINDING_LIGHT = GetSpellInfo(115750)
local DRAGONS_BREATH = GetSpellInfo(31661)
local RING_OF_FROST = GetSpellInfo(113724)
local URSOLS_VORTEX = GetSpellInfo(102793)
local TYPHOON = GetSpellInfo(132469)
local PSYCHIC_SCREAM = GetSpellInfo(8122)
local SHADOWFURY = GetSpellInfo(30283)
local ARCANE_TORRENT = GetSpellInfo(28730)
local WAR_STOMP = GetSpellInfo(20549)

local Backdrop = {
	bgFile = "Interface\\Buttons\\WHITE8X8",
	insets = {top = 0, left = 0, bottom = 0, right = 0},
}

local BackdropAndBorder = {
	bgFile = "Interface\\Buttons\\WHITE8X8",
	edgeFile = "Interface\\Buttons\\WHITE8X8",
	edgeSize = 1,
	insets = {top = 0, left = 0, bottom = 0, right = 0},
}

local StripRealm = function(name)
	if strfind(name, "-") then
		local name = strsplit("-", name)
		
		return name
	else
		return name
	end
end

local OnMouseUp = function(self, button)
	if self.Locked then
		return
	end
	
	if (button == "RightButton") then
		self:ClearAllPoints()
		self:SetPoint("CENTER", UIParent, "CENTER", 0, 120)
	end
end

local OnMouseUpHeader = function(self, button)
	if self.Locked then
		return
	end
	
	if (button == "RightButton") then
		self:ClearAllPoints()
		self:SetPoint("LEFT", UIParent, "LEFT", 8, 0)
	end
end

local HeaderFrame = CreateFrame("Frame", "CCIconsHeaderParent", UIParent)
HeaderFrame:SetPoint("LEFT", UIParent, "LEFT", 8, 0)
HeaderFrame:SetWidth(96)
HeaderFrame:SetHeight(213)
HeaderFrame:SetBackdrop(BackdropAndBorder)
HeaderFrame:SetBackdropColor(0.17, 0.17, 0.17, 0.8)
HeaderFrame:SetBackdropBorderColor(0, 0, 0)
HeaderFrame:Hide()
HeaderFrame:SetMovable(true)
HeaderFrame:SetUserPlaced(true)
HeaderFrame:SetScript("OnDragStart", HeaderFrame.StartMoving)
HeaderFrame:SetScript("OnDragStop", HeaderFrame.StopMovingOrSizing)
HeaderFrame:SetScript("OnMouseUp", OnMouseUpHeader)
HeaderFrame.Locked = true

HeaderFrame.Text = HeaderFrame:CreateFontString(nil, "OVERLAY")
HeaderFrame.Text:SetFont(Font, 14, "OUTLINE")
HeaderFrame.Text:SetPoint("CENTER", HeaderFrame, 0, 0)
HeaderFrame.Text:SetText("Move me!")

-- Dump debuffs on target for CC ID's
-- /run for i = 1, 40 do local n, _, _, _, _, _, _, _, _, _, id = UnitDebuff("target", i); if n then print(n, id) end end

local GroupCDs = {}
local Headers = {}

local Database = {
	-- Death Knight
	[115001] = 6, -- Remorseless Winter
	
	-- Warrior
	[132168] = 4, -- Shockwave
	
	-- Monk
	[119381] = 5, -- Leg Sweep
	[120086] = 3.5, -- Fists of Fury
	
	-- Shaman
	[118905] = 5, -- Static Charge (Capacitor Totem)
	
	-- Druid
	
	-- Rogue
	
	-- Mage
	[31661] = 4, -- Dragon's Breath
	[82691] = 10, -- Ring of Frost
	
	-- Warlock
	[30283] = 3, -- Shadowfury
	
	-- Priest
	[8122] = 8, -- Psychic Scream
	
	-- Paladin
	[105421] = 6, -- Blinding Light
	
	-- Hunter
	[117526] = 5, -- Binding Shot
	
	-- Racials
	[20549] = 2, -- War Stomp
	[25046] = 2, -- Arcane Torrent (Energy)
	[28730] = 2, -- Arcane Torrent (Mana)
	[50613] = 2, -- Arcane Torrent (Runic Power)
	[69179] = 2, -- Arcane Torrent (Rage)
	[80483] = 2, -- Arcane Torrent (Focus)
	[129597] = 2, -- Arcane Torrent (Chi)
	
	-- Professions
	[67890] = 3, -- Cobalt Frag Bomb (Engineering)
}

local Cooldowns = {
	-- Death Knight
	[REMORSELESS_WINTER] = 60,
	
	-- Warrior
	[SHOCKWAVE] = 40,
	
	-- Monk
	[LEG_SWEEP] = 45,
	[FISTS_OF_FURY] = 25,
	
	-- Shaman
	[CAPACITOR_TOTEM] = 45,
	
	-- Mage
	[DRAGONS_BREATH] = 20,
	[RING_OF_FROST] = 45,
	
	-- Hunter
	[BINDING_SHOT] = 45,
	
	-- Druid
	[URSOLS_VORTEX] = 60,
	[TYPHOON] = 30,
	
	-- Priest
	[PSYCHIC_SCREAM] = 30,
	
	-- Warlock
	[SHADOWFURY] = 30,
	
	-- Paladin
	[BLINDING_LIGHT] = 120,
	
	-- Racials
	[ARCANE_TORRENT] = 90,
	[WAR_STOMP] = 90,
}

local SortHeaders = function()
	for i = 1, #Headers do
		Headers[i]:ClearAllPoints()
		
		if (i == 1) then
			Headers[i]:SetPoint("TOP", HeaderFrame, 0, 0)
		else
			Headers[i]:SetPoint("TOP", Headers[i-1], "BOTTOM", 0, -36)
		end
	end
end

local CreateIcon = function(parent)
	local Icon = CreateFrame("Frame", nil, parent)
	Icon:SetPoint("CENTER", UIParent, "CENTER", 0, 120)
	Icon:SetSize(30, 30)
	Icon:SetBackdrop(Backdrop)
	Icon:SetBackdropColor(0, 0, 0)
	
	Icon.Icon = Icon:CreateTexture(nil, "ARTWORK")
	Icon.Icon:SetPoint("TOPLEFT", Icon, 1, -1)
	Icon.Icon:SetPoint("BOTTOMRIGHT", Icon, -1, 1)
	Icon.Icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
	
	Icon.CD = CreateFrame("Cooldown", nil, Icon, "CooldownFrameTemplate")
	Icon.CD:SetPoint("TOPLEFT", Icon, 1, -1)
	Icon.CD:SetPoint("BOTTOMRIGHT", Icon, -1, 1)
	
	local NumRegions = Icon.CD:GetNumRegions()
	
	for i = 1, NumRegions do
		local Region = select(i, Icon.CD:GetRegions())
		
		if Region.GetText then
			Region:SetFont(Font, 14, "OUTLINE")
			Region:SetPoint("CENTER", 1, 0)
			Region:SetTextColor(1, 1, 1)
		end 
	end
	
	return Icon
end

local PlayerHasHeader = function(guid)
	for i = 1, #Headers do
		if (Headers[i].GUID == guid) then
			return true
		end
	end
	
	return false
end

local SortIcons = function(header)
	for i = 1, #header.Icons do
		header.Icons[i]:ClearAllPoints()
	
		if (i == 1) then
			header.Icons[i]:SetPoint("TOPLEFT", header, "BOTTOMLEFT", 0, -3)
		else
			header.Icons[i]:SetPoint("LEFT", header.Icons[i-1], "RIGHT", 3, 0)
		end
	end
end

local RemoveHeader = function(owner)
	for i = 1, #Headers do
		if (Headers[i].Owner == owner) then
			local Header = tremove(Headers, i)
			
			Header:Hide()
			break
		end
	end
	
	SortHeaders()
end

local HideAllHeaders = function()
	local Header
	
	for i = 1, #Headers do
		Header = tremove(Headers, 1)
		
		Header:Hide()
	end
end

local ResetAllCooldowns = function()
	for i = 1, #Headers do
		for x = 1, #Headers[i].Icons do
			Headers[i].Icons[x].CD:SetCooldown(0, 0)
		end
	end
end

-- Process talent/spec changes
local UpdateHeader = function(guid)
	local Header
	local Icon
	
	for i = 1, #Headers do
		if (Headers[i].GUID == guid) then
			Header = Headers[i]
			break
		end
	end
	
	for i = 1, #Header.Icons do
		Icon = tremove(Header.Icons, 1)
		Icon:Hide()
	end
	
	local NumGroupMembers = GetNumGroupMembers()
	local Name
	local Class
	local Race
	local GUID
	
	for i = 1, NumGroupMembers do
		if (UnitGUID("party"..i) == guid) then
			Name = UnitName("party"..i)
			GUID = UnitGUID("party"..i)
			Class = select(2, UnitClass("party"..i))
			Race = UnitRace("party"..i)
			
			break
		end
	end
	
	CreateHeader(Name, GUID, Class, Race)
end

local ClassCDs = {
	["HUNTER"] = function(guid, info, header)
		if info.talents[19347] then -- Binding Shot
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(109248)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][109248] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["DEATHKNIGHT"] = function(guid, info, header)
		if info.talents[19231] then -- Remorseless Winter
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(108200)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][108200] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["WARRIOR"] = function(guid, info, header)
		if info.talents[15760] then -- Shockwave
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(46968)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			Icon.Count = 0 -- Shockwave hit counter
			
			GroupCDs[guid][46968] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["MONK"] = function(guid, info, header)
		if (info.global_spec_id == 269) then -- Windwalking, Fists of Fury
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(113656)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][113656] = Icon
			
			tinsert(header.Icons, Icon)
		end
		
		if info.talents[19995] then -- Leg Sweep
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(119381)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][119381] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["MAGE"] = function(guid, info, header)
		if (info.global_spec_id == 63) then -- Fire, Dragon's Breath
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(31661)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][31661] = Icon
			
			tinsert(header.Icons, Icon)
		end
		
		if info.talents[16019] then -- Ring of Frost
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(113724)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][113724] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["PRIEST"] = function(guid, info, header)
		if info.talents[19768] then -- Psychic Scream
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(8122)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][8122] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["DRUID"] = function(guid, info, header)
		if info.talents[18582] then -- Ursol's Vortex
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(102793)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][102793] = Icon
			
			tinsert(header.Icons, Icon)
		end
		
		if info.talents[18577] then -- Typhoon
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(132469)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][61391] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["WARLOCK"] = function(guid, info, header)
		if info.talents[19286] then -- Shadowfury
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(30283)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][30283] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
	
	["PALADIN"] = function(guid, info, header)
		if info.talents[17577] then -- Blinding Light
			local SpellName, SpellRank, SpellIcon = GetSpellInfo(115750)
			local Icon = CreateIcon(header)
			
			Icon.Icon:SetTexture(SpellIcon)
			Icon.Cooldown = Cooldowns[SpellName]
			
			GroupCDs[guid][115750] = Icon
			
			tinsert(header.Icons, Icon)
		end
	end,
}

local RaceCDs = {
	["Blood Elf"] = function(guid, header)
		local SpellName, SpellRank, SpellIcon = GetSpellInfo(25046) -- Arcane Torrent
		local Icon = CreateIcon(header)
		
		Icon.Icon:SetTexture(SpellIcon)
		Icon.Cooldown = Cooldowns[SpellName]
		
		GroupCDs[guid][ARCANE_TORRENT] = Icon
		
		tinsert(header.Icons, Icon)
	end,
	
	["Tauren"] = function(guid, header)
		local SpellName, SpellRank, SpellIcon = GetSpellInfo(20549) -- War Stomp
		local Icon = CreateIcon(header)
		
		Icon.Icon:SetTexture(SpellIcon)
		Icon.Cooldown = Cooldowns[SpellName]
		
		GroupCDs[guid][20549] = Icon
		
		tinsert(header.Icons, Icon)
	end,
}


local ArcaneTorrent = {
	[25046] = 1,
	[28730] = 1,
	[50613] = 1,
	[69179] = 1,
	[80483] = 1,
	[129597] = 1,
}

local CreateHeader = function(name, guid, class, race)
	if PlayerHasHeader(guid) then
		return
	end
	
	if (not class) then
		return
	end
	
	if (not LGI:GetCachedInfo(guid)) then
		return
	end
	
	local Info = LGI:GetCachedInfo(guid)
	
	local Header = CreateFrame("Frame", nil, UIParent)
	Header:SetSize(96, 18)
	Header:SetBackdrop(BackdropAndBorder)
	Header:SetBackdropColor(0.17, 0.17, 0.17)
	Header:SetBackdropBorderColor(0, 0, 0)
	Header.Owner = name
	Header.GUID = guid
	Header.Icons = {}
	
	Header.Tex = Header:CreateTexture(nil, "OVERLAY")
	Header.Tex:SetTexture(Texture)
	Header.Tex:SetPoint("TOPLEFT", Header, 1, -1)
	Header.Tex:SetPoint("BOTTOMRIGHT", Header, -1, 1)
	Header.Tex:SetVertexColor(0.17, 0.17, 0.17)
	
	if class then
		local Color = RAID_CLASS_COLORS[class].colorStr
		
		if Color then
			name = "|c" .. Color .. StripRealm(name) .. "|r"
		end
	end
	
	Header.Text = Header:CreateFontString(nil, "OVERLAY")
	Header.Text:SetFont(Font, 12, "OUTLINE")
	Header.Text:SetWidth(88)
	Header.Text:SetPoint("LEFT", Header, 2, 0)
	Header.Text:SetJustifyH("LEFT")
	Header.Text:SetText(name)
	
	-- Create icons for user
	if (not GroupCDs[guid]) then
		GroupCDs[guid] = {}
	end
	
	-- Add class cooldowns
	if ClassCDs[class] then
		ClassCDs[class](guid, Info, Header)
	end
	
	-- Add race cooldowns
	if RaceCDs[race] then
		RaceCDs[race](guid, Header)
	end
	
	-- DEBUG! Do a check if this player had anything valuable to show, otherwise dont show it
	tinsert(Headers, Header)
	SortIcons(Header)
	SortHeaders()
	
	return Header
end

local CheckGroup = function()
	local NumGroupMembers = GetNumGroupMembers()
	local Name
	local Class
	local Race
	local GUID
	
	for i = 1, NumGroupMembers do
		Name = UnitName("party"..i)
		GUID = UnitGUID("party"..i)
		Class = select(2, UnitClass("party"..i))
		Race = UnitRace("party"..i)
		
		CreateHeader(Name, GUID, Class, Race)
	end
	
	--[[ DEBUG! PLAYER!
	Name = UnitName("player")
	GUID = UnitGUID("player")
	Class = select(2, UnitClass("player"))
	Race = UnitRace("player")
	
	CreateHeader(Name, GUID, Class, Race)]]
end

local CheckGroupMember = function(unit)
	if UnitExists(unit) then
		local Name = UnitName(unit)
		local GUID = UnitGUID(unit)
		local Class = select(2, UnitClass(unit))
		local Race = UnitRace(unit)
		
		CreateHeader(Name, GUID, Class, Race)
	end
end

local CheckInstance = function(self)
	local InInstance, InstanceType = IsInInstance()
	
	if (InInstance and InstanceType == "party") then
		self:RegisterEvent("CHALLENGE_MODE_RESET")
		self:RegisterEvent("CHALLENGE_MODE_COMPLETED")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		
		LGI:Rescan()
	else
		self:UnregisterEvent("CHALLENGE_MODE_RESET")
		self:UnregisterEvent("CHALLENGE_MODE_COMPLETED")
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		HideAllHeaders()
	end
end

local IconFrame = CreateFrame("Frame", "CCIconsFrame", UIParent)
IconFrame:SetPoint("CENTER", UIParent, "CENTER", 0, 120)
IconFrame:SetWidth(60)
IconFrame:SetHeight(60)
IconFrame:SetBackdrop(Backdrop)
IconFrame:SetBackdropColor(0, 0, 0)
IconFrame:Hide()
IconFrame:SetMovable(true)
IconFrame:SetUserPlaced(true)
IconFrame:SetScript("OnDragStart", IconFrame.StartMoving)
IconFrame:SetScript("OnDragStop", IconFrame.StopMovingOrSizing)
IconFrame:SetScript("OnMouseUp", OnMouseUp)
IconFrame.Locked = true
IconFrame.Alpha = 1

IconFrame.Icon = IconFrame:CreateTexture(nil, "ARTWORK")
IconFrame.Icon:SetPoint("TOPLEFT", IconFrame, 1, -1)
IconFrame.Icon:SetPoint("BOTTOMRIGHT", IconFrame, -1, 1)
IconFrame.Icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

IconFrame.Text = IconFrame:CreateFontString(nil, "OVERLAY")
IconFrame.Text:SetFont(Font, 14, "OUTLINE")
IconFrame.Text:SetPoint("TOP", IconFrame, "BOTTOM", 0, -5)

IconFrame.CD = CreateFrame("Cooldown", "CCIconsCooldown", IconFrame, "CooldownFrameTemplate")
IconFrame.CD:SetPoint("TOPLEFT", IconFrame, 1, -1)
IconFrame.CD:SetPoint("BOTTOMRIGHT", IconFrame, -1, 1)

local NumRegions = IconFrame.CD:GetNumRegions()

for i = 1, NumRegions do
	local Region = select(i, IconFrame.CD:GetRegions())
	
	if Region.GetText then
		Region:SetFont(Font, 18, "OUTLINE")
		Region:SetPoint("CENTER", 1, 0)
		Region:SetTextColor(1, 1, 1)
	end 
end

-- Shockwave CD is -20s if it hits 3 or more targets
local CountShockwave = function(guid)
	if (not GroupCDs[guid][46968]) then
		return
	end
	
	GroupCDs[guid][46968].Count = GroupCDs[guid][46968].Count + 1
	
	C_Timer.After(0.5, function()
		if (GroupCDs[guid][46968].Count >= 3) then
			GroupCDs[guid][46968].CD:SetCooldown(GetTime(), 19.5)
			GroupCDs[guid][46968].Count = 0
		end
	end)
end

local OnUpdate = function(self, elapsed)
	self.Alpha = (self.Alpha - elapsed * 1.8)
	
	self:SetAlpha(self.Alpha)
	
	if (self.Alpha <= 0) then
		self:SetScript("OnUpdate", nil)
		self:Hide()
		self:SetAlpha(1)
		self.Alpha = 1
	end
end

function EventFrame:GroupInfoReady()
	if (not ENABLE_COOLDOWNS) then
		return
	end
	
	--[[local p = LGI:GetCachedInfo(UnitGUID("player"))
	
	if p.talents then
		for k, v in pairs(p.talents) do
			print(k, v.name_localized)
		end
	end]]
	
	local InInstance, InstanceType = IsInInstance()
	
	-- This is crazy long winded and seems like overkill, but sometimes the lib doesn't successfully get talents. So be 100% sure there's talents before making the header.
	if (InInstance and InstanceType == "party") then
		local Party1 = LGI:GetCachedInfo(UnitGUID("party1"))
		local Party2 = LGI:GetCachedInfo(UnitGUID("party2"))
		local Party3 = LGI:GetCachedInfo(UnitGUID("party3"))
		local Party4 = LGI:GetCachedInfo(UnitGUID("party4"))
		local TalentCount = 0
		
		-- P1
		if (Party1 and Party1.talents) then
			for k, v in pairs(Party1.talents) do
				TalentCount = TalentCount + 1
			end
			
			if (TalentCount > 0) then
				CheckGroupMember("party1")
			end
		end
		
		-- P2
		TalentCount = 0
		
		if (Party2 and Party2.talents) then
			for k, v in pairs(Party2.talents) do
				TalentCount = TalentCount + 1
			end
			
			if (TalentCount > 0) then
				CheckGroupMember("party2")
			end
		end
		
		-- P3
		TalentCount = 0
		
		if (Party3 and Party3.talents) then
			for k, v in pairs(Party3.talents) do
				TalentCount = TalentCount + 1
			end
			
			if (TalentCount > 0) then
				CheckGroupMember("party3")
			end
		end
		
		-- P4
		TalentCount = 0
		
		if (Party4 and Party4.talents) then
			for k, v in pairs(Party4.talents) do
				TalentCount = TalentCount + 1
			end
			
			if (TalentCount > 0) then
				CheckGroupMember("party4")
			end
		end
	end
	
	--[[ !DEBUG!
	if LGI:GetCachedInfo(UnitGUID("player")) then
		local Info = LGI:GetCachedInfo(UnitGUID("player"))
		
		for k, v in pairs(Info.talents) do
			print(k, v.name_localized)
		end
	end
	
	CheckGroup()]]
end

function EventFrame:GROUP_ROSTER_UPDATE()
	local NumGroupMembers = GetNumGroupMembers()
	
	if (NumGroupMembers == 0) then
		HideAllHeaders()
	else
		for i = 1, #Headers do
			if Headers[i] then
				if (not UnitInParty(Headers[i].Owner)) then
					RemoveHeader(Headers[i].Owner)
				end
			end
		end
	end
end

function EventFrame:PLAYER_ENTERING_WORLD()
	CheckInstance(self)
end

function EventFrame:CHALLENGE_MODE_RESET()
	if IconFrame:IsVisible() then
		IconFrame:Hide()
		
		if IconFrame:GetScript("OnUpdate") then
			IconFrame:SetScript("OnUpdate", nil)
			IconFrame:SetAlpha(1)
			IconFrame.Alpha = 1
		end
	end
	
	ResetAllCooldowns()
end

function EventFrame:CHALLENGE_MODE_COMPLETED()
	if IconFrame:IsVisible() then
		IconFrame:Hide()
		
		if IconFrame:GetScript("OnUpdate") then
			IconFrame:SetScript("OnUpdate", nil)
			IconFrame:SetAlpha(1)
			IconFrame.Alpha = 1
		end
	end
	
	ResetAllCooldowns()
end

function EventFrame:COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, hide, sourceguid, sourcename, sourceflags, sourcerf, destguid, destname, destflags, destrf, id, name)
	if (event == "SPELL_AURA_APPLIED") and Database[id] and ENABLE_ICONS then
		IconFrame.Icon:SetTexture(select(3, GetSpellInfo(id)))
		IconFrame.Text:SetText(name)
		IconFrame.CD:SetCooldown(GetTime(), Database[id])
		
		if (not IconFrame:IsVisible()) then
			IconFrame:Show()
		end
		
		if IconFrame:GetScript("OnUpdate") then
			IconFrame:SetScript("OnUpdate", nil)
			IconFrame:SetAlpha(1)
			IconFrame.Alpha = 1
		end
		
		if (id == 132168) then
			CountShockwave(sourceguid)
		end
	elseif (event == "SPELL_AURA_REMOVED") and Database[id] then
		if IconFrame:IsVisible() then
			IconFrame:SetScript("OnUpdate", OnUpdate)
		end
	elseif (event == "SPELL_CAST_SUCCESS") and (Cooldowns[name] or ArcaneTorrent[id]) and ENABLE_COOLDOWNS then
		if (GroupCDs[sourceguid] and GroupCDs[sourceguid][id]) then
			GroupCDs[sourceguid][id].CD:SetCooldown(GetTime(), Cooldowns[name])
		elseif (name == ARCANE_TORRENT and GroupCDs[sourceguid] and GroupCDs[sourceguid][ARCANE_TORRENT]) then
			GroupCDs[sourceguid][ARCANE_TORRENT].CD:SetCooldown(GetTime(), Cooldowns[name])
		end
	end
end

function EventFrame:UNIT_SPELLCAST_SUCCEEDED(unit, spell, rank, line, id)
	if (unit ~= "player") then
		return
	end
	
	print(spell, id)
end

function EventFrame:INSPECT_READY(guid)
	
end

function EventFrame:VARIABLES_LOADED()
	if (not CCIconsData) then
		CCIconsData = {ENABLE_ICONS = true, CC_ICON_SIZE = 60, ENABLE_COOLDOWNS = true}
	end
	
	ENABLE_ICONS = CCIconsData.ENABLE_ICONS
	ENABLE_COOLDOWNS = CCIconsData.ENABLE_COOLDOWNS
	CC_ICON_SIZE = CCIconsData.CC_ICON_SIZE or 60
	
	IconFrame:SetSize(CC_ICON_SIZE, CC_ICON_SIZE)
end

EventFrame:RegisterEvent("GROUP_ROSTER_UPDATE")
EventFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
EventFrame:RegisterEvent("VARIABLES_LOADED")
--EventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED") -- DEBUG!
EventFrame:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)

local ToggleMovers = function()
	if IconFrame.Locked then
		IconFrame:EnableMouse(true)
		IconFrame:RegisterForDrag("LeftButton")
		IconFrame.Text:SetText("Move me!")
		IconFrame.Icon:SetTexture("Interface\\ICONS\\inv_misc_bearcubbrown")
		IconFrame:Show()
		IconFrame.Locked = false
		
		HeaderFrame:EnableMouse(true)
		HeaderFrame:RegisterForDrag("LeftButton")
		HeaderFrame:Show()
		HeaderFrame.Locked = false
	else
		IconFrame:EnableMouse(false)
		IconFrame:Hide()
		IconFrame.Text:SetText("")
		IconFrame.Icon:SetTexture("")
		IconFrame.Locked = true
		
		HeaderFrame:EnableMouse(false)
		HeaderFrame:Hide()
		HeaderFrame.Locked = true
	end
end

local EditBoxOnEnterPressed = function(self)
	self:SetAutoFocus(false)
	self:ClearFocus()
	
	local Value = self:GetText()
	
	if (type(tonumber(Value)) == "number") then
		Value = tonumber(Value)
	end
	
	if (Value > 120) then
		Value = 120
	elseif (Value < 20) then
		Value = 20
	end
	
	CC_ICON_SIZE = Value
	CCIconsData.CC_ICON_SIZE = Value
	IconFrame:SetSize(CC_ICON_SIZE, CC_ICON_SIZE)
end

local EditBoxOnMouseDown = function(self)
	self:SetAutoFocus(true)
end

local EditBoxOnEditFocusLost = function(self)
	self:SetAutoFocus(false)
end

local EditBoxOnMouseWheel = function(self, delta)
	local Value = tonumber(self:GetText())
	
	if (delta > 0) then
		Value = Value + 1
		
		if (Value > 120) then
			Value = 120
		end
	else
		Value = Value - 1
		
		if (Value < 20) then
			Value = 20
		end
	end
	
	self:SetText(Value)
	
	CC_ICON_SIZE = Value
	CCIconsData.CC_ICON_SIZE = Value
	IconFrame:SetSize(CC_ICON_SIZE, CC_ICON_SIZE)
end

-- Config GUI
local CreateGUI = function()
	local GUI = CreateFrame("Frame", "CCIGUIFrame", UIParent)
	GUI:SetSize(175, 121)
	GUI:SetPoint("CENTER", UIParent, 0, -30)
	GUI:SetBackdrop(BackdropAndBorder)
	GUI:SetBackdropColor(0.17, 0.17, 0.17)
	GUI:SetBackdropBorderColor(0, 0, 0)
	
	local Header = CreateFrame("Frame", nil, GUI)
	Header:SetSize(175, 22)
	Header:SetPoint("BOTTOMLEFT", GUI, "TOPLEFT", 0, -1)
	Header:SetBackdrop(BackdropAndBorder)
	Header:SetBackdropColor(0.17, 0.17, 0.17)
	Header:SetBackdropBorderColor(0, 0, 0)
	
	Header.Tex = Header:CreateTexture(nil, "OVERLAY")
	Header.Tex:SetTexture(Texture)
	Header.Tex:SetPoint("TOPLEFT", Header, 1, -1)
	Header.Tex:SetPoint("BOTTOMRIGHT", Header, -1, 1)
	Header.Tex:SetVertexColor(0.17, 0.17, 0.17)
	
	Header.Text = Header:CreateFontString(nil, "OVERLAY")
	Header.Text:SetFont(Font, 14, "OUTLINE")
	Header.Text:SetPoint("LEFT", Header, 3, -0.5)
	Header.Text:SetJustifyH("LEFT")
	Header.Text:SetText("CCIcons v1.2")
	
	local Close = CreateFrame("Frame", nil, GUI)
	Close:SetSize(22, 22)
	Close:SetFrameLevel(Header:GetFrameLevel() + 1)
	Close:SetPoint("RIGHT", Header, 0, 0)
	Close:SetBackdrop(BackdropAndBorder)
	Close:SetBackdropColor(0.17, 0.17, 0.17)
	Close:SetBackdropBorderColor(0, 0, 0)
	Close:SetScript("OnEnter", function(self) self.Text:SetTextColor(1, 0, 0) end)
	Close:SetScript("OnLeave", function(self) self.Text:SetTextColor(1, 1, 1) end)
	Close:SetScript("OnMouseUp", function() CCIGUIFrame:Hide() end)
	
	Close.Tex = Close:CreateTexture(nil, "OVERLAY")
	Close.Tex:SetTexture(Texture)
	Close.Tex:SetPoint("TOPLEFT", Close, 1, -1)
	Close.Tex:SetPoint("BOTTOMRIGHT", Close, -1, 1)
	Close.Tex:SetVertexColor(0.17, 0.17, 0.17)
	
	Close.Text = Close:CreateFontString(nil, "OVERLAY")
	Close.Text:SetFont(Font, 16, "OUTLINE")
	Close.Text:SetPoint("CENTER", Close, 0, 1)
	Close.Text:SetJustifyH("LEFT")
	Close.Text:SetText("×")
	
	-- Options
	
	-- Enable stun icons
	local StunEnable = CreateFrame("Frame", nil, GUI)
	StunEnable:SetSize(19, 19)
	StunEnable:SetPoint("TOPLEFT", GUI, 5, -5)
	StunEnable:SetBackdrop(BackdropAndBorder)
	StunEnable:SetBackdropColor(0.17, 0.17, 0.17)
	StunEnable:SetBackdropBorderColor(0, 0, 0)
	StunEnable:SetScript("OnMouseUp", function(self)
		if ENABLE_ICONS then
			ENABLE_ICONS = false
			CCIconsData.ENABLE_ICONS = false
			self.Tex:SetVertexColor(0.8, 0, 0)
		else
			ENABLE_ICONS = true
			CCIconsData.ENABLE_ICONS = true
			self.Tex:SetVertexColor(0, 0.8, 0)
		end
	end)
	
	StunEnable.Tex = StunEnable:CreateTexture(nil, "OVERLAY")
	StunEnable.Tex:SetTexture(Texture)
	StunEnable.Tex:SetPoint("TOPLEFT", StunEnable, 1, -1)
	StunEnable.Tex:SetPoint("BOTTOMRIGHT", StunEnable, -1, 1)
	
	StunEnable.Text = StunEnable:CreateFontString(nil, "OVERLAY")
	StunEnable.Text:SetFont(Font, 12, "OUTLINE")
	StunEnable.Text:SetPoint("LEFT", StunEnable, "RIGHT", 3, 0)
	StunEnable.Text:SetJustifyH("LEFT")
	StunEnable.Text:SetText("Enable stun icons")
	
	-- Enable party cooldowns
	local CDEnable = CreateFrame("Frame", nil, GUI)
	CDEnable:SetSize(19, 19)
	CDEnable:SetPoint("TOPLEFT", StunEnable, "BOTTOMLEFT", 0, -4)
	CDEnable:SetBackdrop(BackdropAndBorder)
	CDEnable:SetBackdropColor(0.17, 0.17, 0.17)
	CDEnable:SetBackdropBorderColor(0, 0, 0)
	CDEnable:SetScript("OnMouseUp", function(self)
		if ENABLE_COOLDOWNS then
			ENABLE_COOLDOWNS = false
			CCIconsData.ENABLE_COOLDOWNS = false
			self.Tex:SetVertexColor(0.8, 0, 0)
		else
			ENABLE_COOLDOWNS = true
			CCIconsData.ENABLE_COOLDOWNS = true
			self.Tex:SetVertexColor(0, 0.8, 0)
		end
	end)
	
	CDEnable.Tex = CDEnable:CreateTexture(nil, "OVERLAY")
	CDEnable.Tex:SetTexture(Texture)
	CDEnable.Tex:SetPoint("TOPLEFT", CDEnable, 1, -1)
	CDEnable.Tex:SetPoint("BOTTOMRIGHT", CDEnable, -1, 1)
	
	CDEnable.Text = CDEnable:CreateFontString(nil, "OVERLAY")
	CDEnable.Text:SetFont(Font, 12, "OUTLINE")
	CDEnable.Text:SetPoint("LEFT", CDEnable, "RIGHT", 3, 0)
	CDEnable.Text:SetJustifyH("LEFT")
	CDEnable.Text:SetText("Enable party cooldowns")
	
	if ENABLE_ICONS then
		StunEnable.Tex:SetVertexColor(0, 0.8, 0)
	else
		StunEnable.Tex:SetVertexColor(0.8, 0, 0)
	end
	
	if ENABLE_COOLDOWNS then
		CDEnable.Tex:SetVertexColor(0, 0.8, 0)
	else
		CDEnable.Tex:SetVertexColor(0.8, 0, 0)
	end
	
	local IconSize = CreateFrame("Frame", nil, GUI)
	IconSize:SetSize(36, 19)
	IconSize:SetPoint("TOPLEFT", CDEnable, "BOTTOMLEFT", 0, -4)
	IconSize:SetBackdrop(BackdropAndBorder)
	IconSize:SetBackdropColor(0.17, 0.17, 0.17)
	IconSize:SetBackdropBorderColor(0, 0, 0)
	
	IconSize.Tex = IconSize:CreateTexture(nil, "OVERLAY")
	IconSize.Tex:SetTexture(Texture)
	IconSize.Tex:SetPoint("TOPLEFT", IconSize, 1, -1)
	IconSize.Tex:SetPoint("BOTTOMRIGHT", IconSize, -1, 1)
	IconSize.Tex:SetVertexColor(0.17, 0.17, 0.17)
	
	IconSize.Text = IconSize:CreateFontString(nil, "OVERLAY")
	IconSize.Text:SetFont(Font, 12, "OUTLINE")
	IconSize.Text:SetPoint("LEFT", IconSize, "RIGHT", 3, 0)
	IconSize.Text:SetJustifyH("LEFT")
	IconSize.Text:SetText("Set icon size")
	
	IconSize.EditBox = CreateFrame("EditBox", nil, IconSize)
	IconSize.EditBox:SetFont(Font, 12, "OUTLINE")
	IconSize.EditBox:SetPoint("TOPLEFT", IconSize, 4, -2)
	IconSize.EditBox:SetPoint("BOTTOMRIGHT", IconSize, -2, 2)
	IconSize.EditBox:SetMaxLetters(3)
	IconSize.EditBox:SetAutoFocus(false)
	IconSize.EditBox:EnableKeyboard(true)
	IconSize.EditBox:EnableMouse(true)
	IconSize.EditBox:EnableMouseWheel(true)
	IconSize.EditBox:SetText(CC_ICON_SIZE)
	IconSize.EditBox:SetScript("OnMouseWheel", EditBoxOnMouseWheel)
	IconSize.EditBox:SetScript("OnMouseDown", EditBoxOnMouseDown)
	IconSize.EditBox:SetScript("OnEscapePressed", EditBoxOnEnterPressed)
	IconSize.EditBox:SetScript("OnEnterPressed", EditBoxOnEnterPressed)
	IconSize.EditBox:SetScript("OnEditFocusLost", EditBoxOnEditFocusLost)
	
	-- Lock/unlock movers
	local MoverButton = CreateFrame("Frame", nil, GUI)
	MoverButton:SetSize(165, 19)
	MoverButton:SetPoint("TOPLEFT", IconSize, "BOTTOMLEFT", 0, -4)
	MoverButton:SetBackdrop(BackdropAndBorder)
	MoverButton:SetBackdropColor(0.17, 0.17, 0.17)
	MoverButton:SetBackdropBorderColor(0, 0, 0)
	MoverButton:SetScript("OnMouseDown", function(self) self.Tex:SetVertexColor(0.2, 0.2, 0.2) end)
	MoverButton:SetScript("OnMouseUp", function(self)
		ToggleMovers()
		self.Tex:SetVertexColor(0.17, 0.17, 0.17)
		
		if IconFrame.Locked then
			self.Text:SetText("Unlock movers")
		else
			self.Text:SetText("Lock movers")
		end
	end)
	
	MoverButton.Tex = MoverButton:CreateTexture(nil, "OVERLAY")
	MoverButton.Tex:SetTexture(Texture)
	MoverButton.Tex:SetPoint("TOPLEFT", MoverButton, 1, -1)
	MoverButton.Tex:SetPoint("BOTTOMRIGHT", MoverButton, -1, 1)
	MoverButton.Tex:SetVertexColor(0.17, 0.17, 0.17)
	
	MoverButton.Text = MoverButton:CreateFontString(nil, "OVERLAY")
	MoverButton.Text:SetFont(Font, 12, "OUTLINE")
	MoverButton.Text:SetPoint("CENTER", MoverButton, 0, 0)
	MoverButton.Text:SetJustifyH("CENTER")
	MoverButton.Text:SetText("Unlock movers")
	
	-- Reload
	local ReloadButton = CreateFrame("Button", nil, GUI)
	ReloadButton:SetSize(165, 19)
	ReloadButton:SetPoint("TOPLEFT", MoverButton, "BOTTOMLEFT", 0, -4)
	ReloadButton:SetBackdrop(BackdropAndBorder)
	ReloadButton:SetBackdropColor(0.17, 0.17, 0.17)
	ReloadButton:SetBackdropBorderColor(0, 0, 0)
	ReloadButton:SetScript("OnClick", ReloadUI)
	
	ReloadButton.Tex = ReloadButton:CreateTexture(nil, "OVERLAY")
	ReloadButton.Tex:SetTexture(Texture)
	ReloadButton.Tex:SetPoint("TOPLEFT", ReloadButton, 1, -1)
	ReloadButton.Tex:SetPoint("BOTTOMRIGHT", ReloadButton, -1, 1)
	ReloadButton.Tex:SetVertexColor(0.17, 0.17, 0.17)
	
	ReloadButton.Text = ReloadButton:CreateFontString(nil, "OVERLAY")
	ReloadButton.Text:SetFont(Font, 12, "OUTLINE")
	ReloadButton.Text:SetPoint("CENTER", ReloadButton, 0, 0)
	ReloadButton.Text:SetJustifyH("CENTER")
	ReloadButton.Text:SetText("Reload UI")
end

local SlashHandler = function(cmd)
	if (cmd == "move") then
		ToggleMovers()
	else
		if (not CCIGUIFrame) then
			return CreateGUI()
		end
		
		if CCIGUIFrame:IsVisible() then
			CCIGUIFrame:Hide()
		else
			CCIGUIFrame:Show()
		end
	end
end

SLASH_CCICONS1 = "/cci"
SlashCmdList["CCICONS"] = SlashHandler

LGI.RegisterCallback(EventFrame, "GroupInSpecT_Update", "GroupInfoReady")